"""RBC FlatBuffer_TextEncodingNone type that corresponds to HeavyDB type
flatbuffer::TextEncodingNone.
"""

__all__ = ["HeavyDBFlatBufferTextEncodingNoneType", "FlatBuffer_TextEncodingNone"]

from .geo_nested_array import (GeoNestedArray, GeoNestedArrayNumbaType,
                               HeavyDBGeoNestedArray)


class FlatBuffer_TextEncodingNoneNumbaType(GeoNestedArrayNumbaType):
    def __init__(self, name):
        super().__init__(name)


class HeavyDBFlatBufferTextEncodingNoneType(HeavyDBGeoNestedArray):
    """Typesystem type class for HeavyDB buffer structures."""

    @property
    def numba_type(self):
        return FlatBuffer_TextEncodingNoneNumbaType

    @property
    def type_name(self):
        return "flatbuffer_TextEncodingNone"

    @property
    def item_type(self):
        return "char8"


class FlatBuffer_TextEncodingNone(GeoNestedArray):
    """
    RBC type that corresponds to HeavyDB type ``TextEncodingNone``

    .. code-block:: c

        struct TextEncodingNone {
            int8_t* flatbuffer_;
            int64_t index_[4];
            int64_t n_;
        }
    """



import pytest

from contextlib import nullcontext as does_not_raise

from rbc.tests import heavydb_fixture
from rbc.typesystem import Type


@pytest.fixture(scope="module")
def heavydb():
    for o in heavydb_fixture(globals(), suffices=[""]):
        yield o


primitives = (
    "void",
    "bool",
    "bool8",
    "int8",
    "int16",
    "int32",
    "int64",
    "float32",
    "float64",
    "short",
    "int",
    "long",
    "long long",
    "float",
    "double",
    # fixed width integer types
    "int8_t",
    "int16_t",
    "int32_t",
    "int64_t",
)

pointers = tuple(map(lambda T: f"{T}*", primitives))

geo = (
    "GeoPoint",
    "GeoMultiPoint",
    "GeoLineString",
    "GeoMultiLineString",
    "GeoPolygon",
    "GeoMultiPolygon",
)
text = ("TextEncodingNone", "TextEncodingDict")
time = ("Timestamp", "DayTimeInterval", "YearMonthTimeInterval")
scalars = primitives + geo + text + time

cursor = tuple(map(lambda T: f"Cursor<{T}>", scalars))
array = tuple(map(lambda T: f"Array<{T}>", scalars))
column = tuple(map(lambda T: f"Column<{T}>", scalars))
column_list = tuple(map(lambda T: f"ColumnList<{T}>", scalars))
column_array = tuple(map(lambda T: f"Column<{T}>", array))
column_list_array = tuple(map(lambda T: f"ColumnList<{T}>", array))

all_types = scalars + pointers + column + column_list + column_array + column_list_array


@pytest.mark.parametrize("device", ("cpu", "gpu"))
@pytest.mark.parametrize("typ", all_types)
def test_type_to_extarg(heavydb, device, typ):
    skiplist = (
        "Cursor",
        "void*",
        "<void>",  # this will cover all containers
        "<DayTimeInterval>",
        "<YearMonthTimeInterval>",
        "ColumnList<Timestamp>",
        "Column<Array<Timestamp>>",
        "ColumnList<Array<Timestamp>>",
        "<Geo",
    )

    for skip in skiplist:
        if skip in typ:
            msg = f"heavydb.type_to_extarg({typ}) not supported."
            pytest.skip(msg)

    if device not in heavydb.targets:
        pytest.skip(f'device "{device}" not available')
    target_info = heavydb.targets[device]
    with target_info:
        with Type.alias(**heavydb.typesystem_aliases):
            t = Type.fromstring(typ)
            with does_not_raise():
                heavydb.type_to_extarg(t)



import pytest

from rbc.tests import heavydb_fixture, override_config


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), load_test_data=False):
        yield o


def test_nrt_warning(heavydb):

    @heavydb("i32(i32)", devices=['cpu'])
    def list_append(t):
        lst = list()
        lst.append('abc')
        return len(lst)

    with override_config('ENABLE_NRT', 0):
        with pytest.warns(UserWarning, match='NRT is disabled but required'):
            heavydb.register()



import pytest
import numpy as np
from rbc.tests import heavydb_fixture
from rbc.stdlib import datatypes, array_api


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), load_test_data=False):
        yield o


@pytest.mark.parametrize('dtype', datatypes.__all__)
def test_datatypes(heavydb, dtype):
    heavydb.unregister()

    @heavydb(f'{dtype}[](int32)')
    def test_datatype(size):
        return array_api.ones(size, dtype=dtype)

    if dtype.startswith('uint') or dtype.startswith('complex'):
        with pytest.raises(ValueError, match=".*cannot convert.*"):
            test_datatype(5).execute()
    else:
        expected = np.ones(5, dtype=dtype)
        got = test_datatype(5).execute()
        np.testing.assert_equal(expected, got)



# tests for how-tos page
# ensure that examples in the docs page are correct
import pytest
from rbc.errors import HeavyDBServerError
from rbc.tests import heavydb_fixture
from rbc.heavydb import RemoteHeavyDB


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals()):
        yield o


def test_connect(heavydb):
    # magictoken.connect.begin
    from rbc.heavydb import RemoteHeavyDB
    heavy = RemoteHeavyDB(user='admin', password='HyperInteractive',
                          host='127.0.0.1', port=6274)
    # magictoken.connect.end
    assert heavy.version is not None


def test_external_functions(heavydb):
    heavydb.unregister()
    # magictoken.external_functions.abs.begin
    from rbc.external import external
    cmath_abs = external('int64 abs(int64)')

    @heavydb('int64(int64)')
    def apply_abs(x):
        return cmath_abs(x)
    # magictoken.external_functions.abs.end

    # magictoken.external_functions.abs.sql.begin
    assert apply_abs(-3).execute() == 3
    # magictoken.external_functions.abs.sql.end

    # magictoken.external_functions.printf.begin
    from rbc.externals.stdio import printf

    @heavydb('int64(int64)')
    def power_2(x):
        # This message will show in the heavydb logs
        printf("input number: %d\n", x)
        return x * x
    # magictoken.external_functions.printf.end

    # magictoken.external_functions.printf.sql.begin
    assert power_2(3).execute() == 9
    # magictoken.external_functions.printf.sql.end


def test_raise_exception(heavydb):
    heavydb.unregister()

    # magictoken.raise_exception.begin
    @heavydb('int32(TableFunctionManager, Column<int64>, OutputColumn<int64>)')
    def udtf_copy(mgr, inp, out):
        size = len(inp)
        if size > 4:
            # error message must be known at compile-time
            return mgr.error_message('TableFunctionManager error_message!')

        mgr.set_output_row_size(size)
        for i in range(size):
            out[i] = inp[i]
        return size
    # magictoken.raise_exception.end

    # magictoken.raise_exception.sql.begin
    query = '''
        SELECT * FROM TABLE(udtf_copy(
            cursor(SELECT * FROM TABLE(generate_series(1, 5, 1)))
        ));
    '''

    with pytest.raises(HeavyDBServerError) as exc:
        heavydb.sql_execute(query)
    exc_msg = ('Error executing table function: TableFunctionManager '
               'error_message!')
    assert exc.match(exc_msg)
    # magictoken.raise_exception.sql.end


def test_devices(heavydb):
    heavydb.unregister()

    # magictoken.devices.begin
    @heavydb('int32(int32, int32)', devices=['CPU', 'GPU'])
    def add(a, b):
        return a + b
    # magictoken.devices.end

    # magictoken.devices.sql.begin
    _, result = heavydb.sql_execute('SELECT add(-3, 3);')
    assert list(result) == [(0,)]
    # magictoken.devices.sql.end


def test_templates(heavydb):
    heavydb.unregister()

    # magictoken.templates.begin
    @heavydb('Z(T, Z)', T=['int32', 'float'], Z=['int64', 'double'])
    def add(a, b):
        return a + b
    # magictoken.templates.end
    heavydb.register()

    # magictoken.templates.sql.begin
    if heavydb.version[:2] >= (6, 2):
        assert heavydb.function_names(runtime_only=True) == ['add']
        assert len(heavydb.function_details('add')) == 4
    assert add(2, 3).execute() == 5
    # magictoken.templates.sql.end


def test_udf(heavydb):
    heavydb.unregister()

    # magictoken.udf.begin
    @heavydb('int64(int64)')
    def incr(a):
        return a + 1
    # magictoken.udf.end

    # magictoken.udf.sql.begin
    _, result = heavydb.sql_execute('SELECT incr(3)')
    assert list(result) == [(4,)]
    # magictoken.udf.sql.end

    heavydb.unregister()

    # magictoken.udf.multiple_signatures.begin
    @heavydb('int64(int64, int64)', 'float64(float64, float64)')
    def multiply(a, b):
        return a * b
    # magictoken.udf.multiple_signatures.end

    # magictoken.udf.multiple_signatures.sql.begin
    _, result = heavydb.sql_execute('SELECT multiply(3.0, 2.0)')
    assert list(result) == [(6.0,)]
    # magictoken.udf.multiple_signatures.sql.end


def test_udtf(heavydb):
    heavydb.unregister()

    # magictoken.udtf.begin
    @heavydb('int32(TableFunctionManager, Column<int64>, OutputColumn<int64>)')
    def my_copy(mgr, inp, out):
        size = len(inp)
        mgr.set_output_row_size(size)
        for i in range(size):
            out[i] = inp[i]
        return size
    # magictoken.udtf.end

    # magictoken.udtf.sql.begin
    query = '''
        SELECT * FROM TABLE(my_copy(
            cursor(SELECT * FROM TABLE(generate_series(1, 5, 1)))
        ));
    '''
    _, result = heavydb.sql_execute(query)
    assert list(result) == [(1,), (2,), (3,), (4,), (5,)]
    # magictoken.udtf.sql.end


def test_udf_text_copy(heavydb):
    heavydb.unregister()
    if heavydb.version[:2] >= (6, 3):
        # magictoken.udf.text.dict.begin
        # Requires HeavyDB server v6.3 or newer
        @heavydb('TextEncodingDict(RowFunctionManager, TextEncodingDict)')
        def text_copy(mgr, t):
            db_id: int = mgr.get_dict_db_id('text_copy', 0)
            dict_id: int = mgr.get_dict_id('text_copy', 0)
            s: str = mgr.get_string(db_id, dict_id, t)
            return mgr.get_or_add_transient(
                mgr.TRANSIENT_DICT_DB_ID,
                mgr.TRANSIENT_DICT_ID,
                s)
        # magictoken.udf.text.dict.end

        table_name = heavydb.table_name + 'text'
        query = f"select text_copy(t1) from {table_name}"
        _, r = heavydb.sql_execute(query)
        assert list(r) == [('fun',), ('bar',), ('foo',), ('barr',), ('foooo',)]


def test_udf_dict_proxy(heavydb):
    heavydb.unregister()
    if heavydb.version[:2] >= (6, 3):
        # magictoken.udf.proxy.begin
        # Requires HeavyDB server v6.3 or newer
        from rbc.heavydb import StringDictionaryProxy

        @heavydb('TextEncodingDict(RowFunctionManager, TextEncodingDict)')
        def test_string_proxy(mgr, t):
            db_id: int = mgr.get_dict_db_id('test_string_proxy', 0)
            dict_id: int = mgr.get_dict_id('test_string_proxy', 0)
            proxy: StringDictionaryProxy = mgr.get_string_dictionary_proxy(db_id, dict_id)
            s: str = proxy.get_string(t)
            return mgr.get_or_add_transient(
                mgr.TRANSIENT_DICT_DB_ID,
                mgr.TRANSIENT_DICT_ID,
                s)
        # magictoken.udf.proxy.end

        table_name = heavydb.table_name + 'text'
        query = f"select test_string_proxy(t1) from {table_name}"
        _, r = heavydb.sql_execute(query)
        assert list(r) == [('fun',), ('bar',), ('foo',), ('barr',), ('foooo',)]


def test_udtf_string_proxy(heavydb):
    heavydb.unregister()

    if heavydb.version[:2] < (6, 2):
        pytest.skip('Test requires HeavyDB 6.1 or newer')

    # magictoken.udtf.proxy.begin
    @heavydb('int32(TableFunctionManager, Column<T>, OutputColumn<T> | input_id=args<0>)',
             T=['TextEncodingDict'])
    def test_string_proxy(mgr, inp, out):
        size = len(inp)
        mgr.set_output_row_size(size)
        for i in range(size):
            s = inp.string_dict_proxy.get_string(inp[i])
            id = out.string_dict_proxy.get_or_add_transient(s.title())
            out[i] = id
        return size
    # magictoken.udtf.proxy.end

    table_name = heavydb.table_name + 'text'
    query = f'''
        SELECT * FROM TABLE(test_string_proxy(
            cursor(SELECT t1 from {table_name})
        ))
    '''
    _, r = heavydb.sql_execute(query)
    assert list(r) == [('Fun',), ('Bar',), ('Foo',), ('Barr',), ('Foooo',)]


def test_udf_text_duplicate(heavydb):
    heavydb.unregister()
    # magictoken.udf.text.none.begin
    from rbc.heavydb import TextEncodingNone

    @heavydb('TextEncodingNone(TextEncodingNone)')
    def text_duplicate(t):
        s: str = t.to_string()
        return TextEncodingNone(s + s)
    # magictoken.udf.text.none.end

    table_name = heavydb.table_name + 'text'
    query = f"select text_duplicate(n) from {table_name}"
    _, r = heavydb.sql_execute(query)
    assert list(r) == [('funfun',), ('barbar',), ('foofoo',), ('barrbarr',),
                       ('foooofoooo',)]


def test_udf_text_capitalize(heavydb):
    heavydb.unregister()
    # magictoken.udf.text.capitalize.begin
    from rbc.heavydb import TextEncodingNone

    @heavydb('TextEncodingNone(TextEncodingNone)')
    def text_capitalize(t):
        s: str = t.to_string()
        return TextEncodingNone(s.capitalize())
    # magictoken.udf.text.capitalize.end

    table_name = heavydb.table_name + 'text'
    query = f"select text_capitalize(n) from {table_name}"
    _, r = heavydb.sql_execute(query)
    assert list(r) == [('Fun',), ('Bar',), ('Foo',), ('Barr',),
                       ('Foooo',)]


def test_array(heavydb):
    heavydb.unregister()
    table_name = heavydb.table_name + 'array'

    # magictoken.udf.array.new.begin
    from rbc.heavydb import Array

    @heavydb('Array<int64>(int32)')
    def arr_new(size):
        arr = Array(size, dtype='int64')
        for i in range(size):
            arr[i] = 1
        return arr
    # magictoken.udf.array.new.end

    # magictoken.udf.array.new.sql.begin
    _, r = heavydb.sql_execute('SELECT arr_new(5);')
    assert list(r) == [([1, 1, 1, 1, 1],)]
    # magictoken.udf.array.new.sql.end

    # magictoken.udf.array.length.begin
    @heavydb('int64(Array<int32>)')
    def my_length(arr):
        return len(arr)
    # magictoken.udf.array.length.end

    # magictoken.udf.array.length.sql.begin
    _, r = heavydb.sql_execute(f'SELECT my_length(i4) from {table_name}')
    assert list(r) == [(0,), (1,), (2,), (3,), (4,)]
    # magictoken.udf.array.length.sql.end

    # magictoken.udf.array.array_api.begin
    from rbc.stdlib import array_api

    @heavydb('Array<int64>(int32)')
    def arr_new_ones(sz):
        return array_api.ones(sz, dtype='int64')
    # magictoken.udf.array.array_api.end

    # magictoken.udf.array.array_api.sql.begin
    _, r = heavydb.sql_execute('SELECT arr_new_ones(5);')
    assert list(r) == [([1, 1, 1, 1, 1],)]
    # magictoken.udf.array.array_api.sql.end


def test_tablefunctionmanager(heavydb):
    heavydb.unregister()

    # magictoken.udtf.mgr.basic.begin
    @heavydb('int32(TableFunctionManager, Column<int64>, OutputColumn<int64>)')
    def table_copy(mgr, inp, out):
        size = len(inp)
        mgr.set_output_row_size(size)
        for i in range(size):
            out[i] = inp[i]
        return size
    # magictoken.udtf.mgr.basic.end

    # magictoken.udtf.mgr.basic.sql.begin
    query = '''
        SELECT * FROM TABLE(table_copy(
            cursor(SELECT * FROM TABLE(generate_series(0, 4, 1)))
        ))
    '''
    _, r = heavydb.sql_execute(query)
    assert list(r) == [(0,), (1,), (2,), (3,), (4,)]
    # magictoken.udtf.mgr.basic.sql.end


def test_rowfunctionmanager(heavydb):
    heavydb.unregister()
    table = heavydb.table_name + 'text'

    if heavydb.version[:2] >= (6, 4):
        # magictoken.udf.mgr.basic.begin
        @heavydb('TextEncodingDict(RowFunctionManager, TextEncodingDict)')
        def concat(mgr, text):
            db_id: int = mgr.get_dict_db_id('concat', 0)
            dict_id: int = mgr.get_dict_id('concat', 0)
            s: str = mgr.get_string(db_id, dict_id, text)
            s_concat = 'test: ' + s
            return mgr.get_or_add_transient(mgr.TRANSIENT_DICT_DB_ID,
                                            mgr.TRANSIENT_DICT_ID,
                                            s_concat)
        # magictoken.udf.mgr.basic.end

        # magictoken.udf.mgr.basic.sql.begin
        query = f'''
            SELECT concat(t1) FROM {table}
        '''
        _, r = heavydb.sql_execute(query)
        assert list(r) == [('test: fun',), ('test: bar',), ('test: foo',),
                           ('test: barr',), ('test: foooo',)]
        # magictoken.udf.mgr.basic.sql.end


def test_column_power(heavydb):
    heavydb.unregister()

    # magictoken.udtf.column.basic.begin
    import numpy as np

    @heavydb('int32(TableFunctionManager, Column<T>, T, OutputColumn<T>)',
             T=['int64'])
    def udtf_power(mgr, inp, exp, out):
        size = len(inp)
        mgr.set_output_row_size(size)
        for i in range(size):
            out[i] = np.power(inp[i], exp)
        return size
    # magictoken.udtf.column.basic.end

    # magictoken.udtf.column.basic.sql.begin
    query = '''
        SELECT * FROM TABLE(udtf_power(
            cursor(SELECT * FROM TABLE(generate_series(1, 5, 1))),
            3
        ))
    '''
    _, r = heavydb.sql_execute(query)
    assert list(r) == [(1,), (8,), (27,), (64,), (125,)]
    # magictoken.udtf.column.basic.sql.end


def test_geopoint(heavydb: RemoteHeavyDB):
    if heavydb.version[:2] < (7, 0):
        pytest.skip('Requires HeavyDB 7.0 or newer')

    heavydb.unregister()

    # magictoken.udtf.geopoint.basic.begin
    from rbc.heavydb import Point2D

    @heavydb('int32(TableFunctionManager, int64 size, OutputColumn<Z>)',
             Z=['GeoPoint'])
    def generate_geo(mgr, size, out):
        mgr.set_output_row_size(size)
        for i in range(size):
            out[i] = Point2D(i, i)
        return size
    # magictoken.udtf.geopoint.basic.end

    # magictoken.udtf.geopoint.basic.sql.begin
    query = '''
        SELECT * FROM TABLE(
            generate_geo(5)
        );
    '''
    _, r = heavydb.sql_execute(query)
    r = list(r)
    assert r == [('POINT (0 0)',), ('POINT (1 1)',), ('POINT (2 2)',),
                 ('POINT (3 3)',), ('POINT (4 4)',)]
    # magictoken.udtf.geopoint.basic.sql.end


def test_geomultipoint(heavydb: RemoteHeavyDB):
    if heavydb.version[:2] < (7, 0):
        pytest.skip('Requires HeavyDB 7.0 or newer')

    heavydb.unregister()

    # magictoken.udtf.mp.basic.begin
    @heavydb('int32(TableFunctionManager, int64 size, OutputColumn<Z>)',
             Z=['GeoMultiPoint'])
    def generate_geo(mgr, size, out):
        mgr.set_output_item_values_total_number(0, size * 4)
        mgr.set_output_row_size(size)
        for i in range(size):
            coords = [i + 1.0, i + 2.0, i + 3.0, i + 4.0]
            out[i].from_coords(coords)
        return size
    # magictoken.udtf.mp.basic.end

    # magictoken.udtf.mp.basic.sql.begin
    query = '''
        SELECT * FROM TABLE(
            generate_geo(3)
        );
    '''
    _, r = heavydb.sql_execute(query)
    r = list(r)
    assert r == [('MULTIPOINT (1 2,3 4)',), ('MULTIPOINT (2 3,4 5)',),
                 ('MULTIPOINT (3 4,5 6)',)]
    # magictoken.udtf.mp.basic.sql.end



"""Implement Heavydb Column<TextEncodingNone> type support

Heavydb Column<TextEncodingNone> type is the type of input/output column arguments in
UDTFs.
"""

__all__ = [
    "HeavyDBOutputColumnTextEncodingNoneType",
    "HeavyDBColumnTextEncodingNoneType",
    "ColumnTextEncodingNone",
]

import operator

from llvmlite import ir
from numba.core import cgutils, extending

from rbc import typesystem
from rbc.external import external

from .buffer import BufferPointer
from .column_flatbuffer import (ColumnFlatBuffer, ColumnFlatBufferPointer,
                                ColumnFlatBufferType,
                                HeavyDBColumnFlatBufferType,
                                HeavyDBOutputColumnFlatBufferType)
from .text_encoding_none import TextEncodingNone
from .utils import as_voidptr, get_alloca

i1 = ir.IntType(1)
i8 = ir.IntType(8)
i8p = i8.as_pointer()
i8pp = i8p.as_pointer()
i32 = ir.IntType(32)
i64 = ir.IntType(64)
i64p = i64.as_pointer()
void = ir.VoidType()


class ColumnTextEncodingNone(ColumnFlatBuffer):
    """
    RBC ``Column<TextEncodingNone>`` type that corresponds to HeavyDB
    ``COLUMN<TEXT_ENCODING_NONE>``

    In HeavyDB, a Column of type ``TextEncodingNone`` is represented as follows:

    .. code-block:: c

        {
            int8_t* flatbuffer;
            int64_t size;
        }

    """

    def __getitem__(self, index: int) -> "TextEncodingNone.TextEncodingNone":
        """
        Return ``self[index]``

        .. note::
            Only available on ``CPU``
        """

    def get_item(self, index: int) -> "TextEncodingNone.TextEncodingNone":
        """
        Return ``self[index]``

        .. note::
            Only available on ``CPU``
        """

    def set_item(self, index: int, buf: "TextEncodingNone.TextEncodingNone") -> None:
        """
        Set line from a buffer of point coordindates

        .. note::
            Only available on ``CPU``
        """

    def __setitem__(self, index: int, buf: "TextEncodingNone.TextEncodingNone") -> None:
        """
        Set line from a buffer of point coordindates

        .. note::
            Only available on ``CPU``
        """


class HeavyDBColumnTextEncodingNoneType(HeavyDBColumnFlatBufferType):
    """ """

    @property
    def numba_type(self):
        return ColumnTextEncodingNoneType

    @property
    def numba_pointer_type(self):
        return ColumnTextEncodingNonePointer

    @property
    def type_name(self):
        return "TextEncodingNone"


class HeavyDBOutputColumnTextEncodingNoneType(
    HeavyDBColumnTextEncodingNoneType, HeavyDBOutputColumnFlatBufferType
):
    """Heavydb OutputColumn type for RBC typesystem.

    OutputColumn<TextEncodingNone> is the same as Column<TextEncodingNone> but introduced to
    distinguish the input and output arguments of UDTFs.
    """


class ColumnTextEncodingNoneType(ColumnFlatBufferType):
    pass


class ColumnTextEncodingNonePointer(ColumnFlatBufferPointer):
    pass


@extending.intrinsic
def heavydb_column_getitem_(typingctx, col, index):
    text = typesystem.Type.fromstring("flatbuffer_TextEncodingNone").tonumba()
    sig = text(col, index)

    def codegen(context, builder, sig, args):
        col, index = args
        fnty = ir.FunctionType(void, [i8p, i64, i8p])
        getItem = cgutils.get_or_insert_function(
            builder.module, fnty, "ColumnTextEncodingNone_getItem"
        )
        col_ptr = builder.bitcast(col, i8p)

        # Alloca TextEncodingNone
        fa = context.make_helper(builder, sig.return_type)
        fa.n_ = fa.n_.type(1)

        # TextEncodingNone -> void*
        result_ptr = builder.bitcast(fa._getpointer(), i8p)

        # call func
        builder.call(getItem, [col_ptr, index, result_ptr])

        # convert void* -> TextEncodingNone
        point_type = fa._get_be_type(fa._datamodel)
        return builder.load(builder.bitcast(result_ptr, point_type.as_pointer()))

    return sig, codegen


@extending.overload(operator.getitem)
@extending.overload_method(ColumnTextEncodingNonePointer, "get_item")
def heavydb_column_getitem(col, index):
    if isinstance(col, ColumnTextEncodingNonePointer):

        def impl(col, index):
            return heavydb_column_getitem_(col, index)

        return impl


@extending.overload(operator.setitem)
@extending.overload_method(ColumnTextEncodingNonePointer, "set_item")
def heavydb_column_set_item(col, index, rhs):
    if isinstance(col, ColumnFlatBufferPointer):
        fn_name = 'ColumnTextEncodingNone_setItem'
        if isinstance(rhs, BufferPointer):
            fn_name += '_fromBuffer'

        sig = f"void {fn_name}(int8_t*, int64_t, int8_t*)"
        setItem = external(sig)

        if isinstance(rhs, BufferPointer):
            def impl(col, index, rhs):
                setItem(as_voidptr(col), index, as_voidptr(rhs))
        else:
            def impl(col, index, rhs):
                setItem(as_voidptr(col), index, as_voidptr(get_alloca(rhs)))

        return impl


@extending.overload_method(ColumnTextEncodingNonePointer, "concat_item")
def heavydb_column_concat_item(col, index, rhs):
    fn_name = 'ColumnTextEncodingNone_concatItem'
    if isinstance(rhs, BufferPointer):
        fn_name += '_fromBuffer'

    sig = f"void {fn_name}(int8_t*, int64_t, int8_t*)"
    concatItem = external(sig)

    if isinstance(col, ColumnFlatBufferPointer):
        if isinstance(rhs, BufferPointer):
            # TextEncodingNone
            def impl(col, index, rhs):
                concatItem(as_voidptr(col), index, as_voidptr(rhs))
        else:
            # flatbuffer::TextEncodingNone
            def impl(col, index, rhs):
                concatItem(as_voidptr(col),
                           index,
                           as_voidptr(get_alloca(rhs)))
        return impl



from numba.core import ir
from numba.core.compiler import CompilerBase, DefaultPassBuilder
from numba.core.compiler_machinery import FunctionPass, register_pass
from numba.core.untyped_passes import IRProcessing

from rbc.errors import NumbaTypeError


@register_pass(mutates_CFG=False, analysis_only=False)
class CheckRaiseStmts(FunctionPass):
    _name = "check_raise_stmts"

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        func_ir = state.func_ir
        for blk in func_ir.blocks.values():
            for _raise in blk.find_insts(ir.Raise):
                msg = ('raise statement is not supported in '
                       'UDF/UDTFs. Please, use `return table_function_error(msg)` '
                       'to raise an error.')
                loc = _raise.loc
                raise NumbaTypeError(msg, loc=loc)
        return False


class HeavyDBCompilerPipeline(CompilerBase):
    def define_pipelines(self):
        # define a new set of pipelines (just one in this case) and for ease
        # base it on an existing pipeline from the DefaultPassBuilder,
        # namely the "nopython" pipeline
        pm = DefaultPassBuilder.define_nopython_pipeline(self.state)
        # Add the new pass to run after IRProcessing
        pm.add_pass_after(CheckRaiseStmts, IRProcessing)
        # finalize
        pm.finalize()
        # return as an iterable, any number of pipelines may be defined!
        return [pm]



import os
import sys
import versioneer

if sys.version_info[:2] < (3, 7):
    raise RuntimeError("Python version >= 3.7 required.")

if os.path.exists('MANIFEST'):
    os.remove('MANIFEST')

CONDA_BUILD = int(os.environ.get('CONDA_BUILD', '0'))
CONDA_ENV = os.environ.get('CONDA_PREFIX', '') != ''

from setuptools import setup, find_packages  # noqa: E402

DESCRIPTION = "RBC - Remote Backend Compiler Project"
LONG_DESCRIPTION = """
The aim of the Remote Backend Compiler project is to distribute the
tasks of a program JIT compilation process to separate computer
systems using the client-server model. The frontend of the compiler
runs on the client computer and the backend runs on the server
computer. The compiler frontend will send the program code to compiler
backend in IR form where it will be compiled to machine code.
"""


def setup_package():
    src_path = os.path.dirname(os.path.abspath(sys.argv[0]))
    old_path = os.getcwd()
    os.chdir(src_path)
    sys.path.insert(0, src_path)

    if CONDA_BUILD or CONDA_ENV:
        # conda dependencies are specified in meta.yaml or conda
        # enviroment should provide the correct requirements - using
        # PyPI is unreliable, see below.
        install_requires = []
        tests_require = []
    else:
        # Get requirements via PyPI. Use at your own risk as more than
        # once the numba and llvmlite have not matched.
        install_requires = open('requirements.txt', 'r').read().splitlines()
        tests_require = ['pytest', 'packaging', 'clang==11.1.0']

    metadata = dict(
        name='rbc-project',
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        license='BSD',
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        author='Pearu Peterson',
        maintainer='Pearu Peterson',
        author_email='pearu.peterson@quansight.com',
        url='https://github.com/xnd-project/rbc',
        platforms='Cross Platform',
        classifiers=[
            "Intended Audience :: Developers",
            "License :: OSI Approved :: BSD License",
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            "Operating System :: OS Independent",
            "Topic :: Software Development",
        ],
        packages=find_packages(),
        package_data={'': ['*.thrift', 'unicodetype_db.ll']},
        install_requires=install_requires,
        tests_require=tests_require,
    )

    try:
        setup(**metadata)
    finally:
        del sys.path[0]
        os.chdir(old_path)
    return


if __name__ == '__main__':
    setup_package()



import numpy as np
import pytest

from rbc.tests import heavydb_fixture

rbc_heavydb = pytest.importorskip('rbc.heavydb')
available_version, reason = rbc_heavydb.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), suffices=['text']):
        define(o)
        yield o


def define(heavydb):

    @heavydb('int32(TableFunctionManager, Column<T>, OutputColumn<T>)',
             T=['TextEncodingNone'])
    def rbc_ct_copy(mgr, inputs, outputs):
        size = len(inputs)
        mgr.set_output_item_values_total_number(0, inputs.get_n_of_values())
        mgr.set_output_row_size(size)
        for i in range(size):
            if inputs.is_null(i):
                outputs.set_null(i)
            else:
                outputs[i] = inputs[i]
        return size

    @heavydb('int32(TableFunctionManager, Cursor<Column<T>, Column<T>>, OutputColumn<T>)',
             T=['TextEncodingNone'], devices=['CPU'])
    def rbc_ct_concat__generic1(mgr, input1, input2, outputs):
        size = len(input1)
        mgr.set_output_item_values_total_number(
            0, input1.get_n_of_values() + input2.get_n_of_values())
        mgr.set_output_row_size(size)
        for i in range(size):
            if input1.is_null(i):
                if input2.is_null(i):
                    outputs.set_null(i)
                else:
                    outputs[i] = input2[i]
            elif input2.is_null(i):
                outputs[i] = input1[i]
            else:
                outputs[i] = input1[i]
                outputs.concat_item(i, input2[i])
        return size

    @heavydb('int32(TableFunctionManager, Column<T>, T, OutputColumn<T>)',
             T=['TextEncodingNone'], devices=['CPU'])
    def rbc_ct_concat__generic2(mgr, input1, input2, output):
        size = len(input1)
        mgr.set_output_item_values_total_number(
            0, input1.get_n_of_values() + size * len(input2))
        mgr.set_output_row_size(size)
        for i in range(size):
            if input1.is_null(i) or input2.is_null():
                output.set_null(i)
            else:
                output[i] = input1[i]
                output.concat_item(i, input2)
        return size

    @heavydb('int32(TableFunctionManager, T, Column<T>, OutputColumn<T>)',
             T=['TextEncodingNone'], devices=['CPU'])
    def rbc_ct_concat__generic3(mgr, input1, input2, output):
        size = len(input2)
        mgr.set_output_item_values_total_number(
            0, input2.get_n_of_values() + size * len(input1))
        mgr.set_output_row_size(size)
        for i in range(size):
            if input2.is_null(i) or input1.is_null():
                output.set_null(i)
            else:
                output[i] = input1
                output.concat_item(i, input2[i])
        return size


@pytest.mark.parametrize('col', ('n', 'n2'))
def test_rbc_ct_copy(heavydb, col):
    if heavydb.version[:2] < (7, 0):
        pytest.skip('Requires HeavyDB 7.0 or newer')

    suffix = 'text'
    table = heavydb.table_name + suffix

    query = (f"SELECT * FROM TABLE(rbc_ct_copy(CURSOR(SELECT "
             f"{col} from {table})));")
    _, result = heavydb.sql_execute(query)

    query = (f"SELECT * FROM TABLE(ct_copy(CURSOR(SELECT "
             f"{col} from {table})));")
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)


@pytest.mark.parametrize('col', ('n', 'n2'))
def test_rbc_ct_concat(heavydb, col):
    if heavydb.version[:2] < (7, 0):
        pytest.skip('Requires HeavyDB 7.0 or newer')

    suffix = 'text'
    table = heavydb.table_name + suffix

    query = (f"SELECT * FROM TABLE(rbc_ct_concat(CURSOR(SELECT "
             f"{col}, {col} from {table})));")
    _, result = heavydb.sql_execute(query)

    query = (f"SELECT * FROM TABLE(ct_concat(CURSOR(SELECT "
             f"{col}, {col} from {table})));")
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)


@pytest.mark.parametrize('col', ('n', 'n2'))
def test_rbc_ct_concat2(heavydb, col):
    if heavydb.version[:2] < (7, 0):
        pytest.skip('Requires HeavyDB 7.0 or newer')

    suffix = 'text'
    var = 'test'
    table = heavydb.table_name + suffix

    query = (f"SELECT * FROM TABLE(rbc_ct_concat(CURSOR(SELECT "
             f"{col} from {table}), '{var}'));")
    _, result = heavydb.sql_execute(query)

    query = (f"SELECT * FROM TABLE(ct_concat(CURSOR(SELECT "
             f"{col} from {table}), '{var}'));")
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)


@pytest.mark.parametrize('col', ('n', 'n2'))
def test_rbc_ct_concat3(heavydb, col):
    if heavydb.version[:2] < (7, 0):
        pytest.skip('Requires HeavyDB 7.0 or newer')

    suffix = 'text'
    var = 'test'
    table = heavydb.table_name + suffix

    query = (f"SELECT * FROM TABLE(rbc_ct_concat('{var}', "
             f"CURSOR(SELECT {col} from {table})));")
    _, result = heavydb.sql_execute(query)

    query = (f"SELECT * FROM TABLE(ct_concat('{var}', "
             f"CURSOR(SELECT {col} from {table})));")
    _, expected = heavydb.sql_execute(query)

    result = list(result)
    expected = list(expected)
    np.testing.assert_equal(result, expected)



"""
Array API specification for element-wise functions.

https://data-apis.org/array-api/latest/API_specification/elementwise_functions.html.
"""

from rbc.stdlib import (
    Expose,
    BinaryUfuncExpose,
    UnaryUfuncExpose,
    API,
    determine_input_type,
)
import numpy as np
from rbc import typesystem
from rbc.heavydb import ArrayPointer, Array
from numba.core import types


__all__ = [
    'add', 'subtract', 'multiply', 'divide', 'logaddexp', 'logaddexp2',
    'true_divide', 'floor_divide', 'pow', 'remainder', 'mod',
    'fmod', 'gcd', 'lcm', 'bitwise_and', 'bitwise_or', 'bitwise_xor',
    'bitwise_invert', 'atan2', 'hypot',
    'greater', 'greater_equal', 'less', 'less_equal', 'not_equal',
    'equal', 'logical_and', 'logical_or', 'logical_xor', 'maximum',
    'minimum', 'fmax', 'fmin', 'nextafter', 'ldexp', 'negative',
    'positive', 'rint', 'sign', 'abs', 'conj',
    'conjugate', 'exp', 'exp2', 'log', 'log2', 'log10', 'expm1',
    'log1p', 'sqrt', 'square', 'reciprocal', 'sin', 'cos',
    'tan', 'asin', 'acos', 'atan', 'sinh', 'cosh', 'tanh',
    'asinh', 'acosh', 'atanh', 'degrees', 'radians', 'deg2rad',
    'rad2deg', 'logical_not', 'isfinite', 'isinf', 'isnan', 'fabs',
    'floor', 'ceil', 'trunc', 'signbit', 'copysign', 'spacing',
    'heaviside', 'bitwise_left_shift', 'bitwise_right_shift',
    'round', 'isnat',
    # numpy specifics
    'power', 'arctan2', 'left_shift', 'right_shift', 'absolute',
    'invert', 'arcsin', 'arctan', 'arccos', 'arcsinh', 'arccosh', 'arctanh',
    'float_power', 'divmod', 'cbrt',
]


expose = Expose(globals(), "elementwise_functions")
binary_expose = BinaryUfuncExpose(globals(), "elementwise_functions")
unary_expose = UnaryUfuncExpose(globals(), "elementwise_functions")


@unary_expose.implements(np.absolute, func_name="abs")
def _impl_ufunc_abs(x):
    """
    Calculates the absolute value for each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_abs(arr):
    ...     return array_api.abs(arr)
    >>> array_api_abs([-1.0, 2.0, -3.0]).execute()
    array([1., 2., 3.], dtype=float32)

    """
    pass


@unary_expose.implements(np.arccos, func_name="acos")
def _impl_ufunc_acos(x):
    """
    Calculates an implementation-dependent approximation of the principal value
    of the inverse cosine for each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_acos(arr):
    ...     return array_api.acos(arr)
    >>> array_api_acos([1., -1.]).execute()
    array([0.       , 3.1415927], dtype=float32)

    """
    pass


@unary_expose.implements(np.arccosh, func_name="acosh")
def _impl_ufunc_acosh(x):
    """
    Calculates an implementation-dependent approximation to the inverse
    hyperbolic cosine for each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_acosh(arr):
    ...     return array_api.acosh(arr)
    >>> array_api_acosh([np.e, 10.0, 1.0]).execute()
    array([1.6574545, 2.993223 , 0.       ], dtype=float32)

    """
    pass


@binary_expose.implements(np.add, func_name="add")
def _impl_add(x1, x2):
    """
    Calculates the sum for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('int64[](int64[], int64[])')
    ... def array_api_add(x1, x2):
    ...     return array_api.add(x1, x2)
    >>> array_api_add([1, -1], [3, 4]).execute()
    array([4, 3])
    """
    pass


@unary_expose.implements(np.arcsin, func_name="asin")
def _impl_ufunc_asin(x):
    """
    Calculates an implementation-dependent approximation of the principal value
    of the inverse sine for each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_asin(arr):
    ...     return array_api.asin(arr)
    >>> array_api_asin([1.0, 0.0, -1.0]).execute()
    array([ 1.5707964,  0.       , -1.5707964], dtype=float32)

    """
    pass


@unary_expose.implements(np.arcsinh, func_name="asinh")
def _impl_ufunc_asinh(x):
    """
    Calculates an implementation-dependent approximation to the inverse
    hyperbolic sine for each element x_i in the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_asinh(arr):
    ...     return array_api.asinh(arr)
    >>> array_api_asinh([np.e, 10.0]).execute()
    array([1.7253826, 2.998223 ], dtype=float32)
    """
    pass


@unary_expose.implements(np.arctan, func_name="atan")
def _impl_ufunc_atan(x):
    """
    Calculates an implementation-dependent approximation of the principal value
    of the inverse tangent for each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_atan(arr):
    ...     return array_api.atan(arr)
    >>> array_api_atan([0, 1, np.pi/4]).execute()
    array([0.        , 0.7853982 , 0.66577375], dtype=float32)

    """
    pass


@binary_expose.implements(np.arctan2, func_name="atan2", dtype=types.float64)
def _impl_ufunc_atan2(x1, x2):
    """
    Calculates an implementation-dependent approximation of the inverse tangent of the quotient
    x1/x2, having domain [-infinity, +infinity] x ``[-infinity, +infinity]`` (where the x notation
    denotes the set of ordered pairs of elements (x1_i, x2_i)) and codomain [-π, +π], for each
    pair of elements (x1_i, x2_i) of the input arrays x1 and x2, respectively.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](int64[], int64[])')
    ... def array_api_atan2(x1, x2):
    ...     return array_api.atan2(x1, x2)
    >>> x = np.array([-1, +1, +1, -1])
    >>> y = np.array([-1, -1, +1, +1])
    >>> array_api_atan2(x, y).execute() * 180 / np.pi
    array([-135.,  135.,   45.,  -45.], dtype=float32)

    """
    pass


@unary_expose.implements(np.arctanh, func_name="atanh")
def _impl_ufunc_atanh(x):
    """
    Calculates an implementation-dependent approximation to the inverse hyperbolic tangent, having
    domain [-1, +1] and codomain [-infinity, +infinity], for each element x_i of the input array
    x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_atanh(x):
    ...     return array_api.atanh(x)
    >>> array_api_atanh([0, -0.5]).execute()
    array([ 0.        , -0.54930615], dtype=float32)

    """
    pass


@binary_expose.implements(np.bitwise_and, func_name="bitwise_and")
def _impl_ufunc_bitwise_and(x1, x2):
    """
    Computes the bitwise AND of the underlying binary representation of each element x1_iof the
    input array x1 with the respective element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('int64[](int64[], int64[])')
    ... def array_api_bitwise_and(x1, x2):
    ...     return array_api.bitwise_and(x1, x2)
    >>> array_api_bitwise_and([11,7], [4,25]).execute()
    array([0, 1])

    """
    pass


@binary_expose.implements(np.left_shift, func_name="bitwise_left_shift")
def _impl_ufunc_bitwise_left_shift(x1, x2):
    """
    Shifts the bits of each element x1_i of the input array x1 to the left by appending x2_i
    (i.e., the respective element in the input array x2) zeros to the right of x1_i.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('int64[](int64, int64[])')
    ... def array_api_bitwise_left_shift(x1, x2):
    ...     return array_api.bitwise_left_shift(x1, x2)
    >>> array_api_bitwise_left_shift(5, [1, 2, 3]).execute()
    array([10, 20, 40])

    """
    pass


@unary_expose.implements(np.invert, func_name="bitwise_invert")
def _impl_ufunc_bitwise_invert(x):
    """
    Inverts (flips) each bit for each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('int8(int8)')
    ... def array_api_bitwise_invert(x):
    ...     return array_api.bitwise_invert(x)
    >>> array_api_bitwise_invert(13).execute()
    -14
    >>> np.binary_repr(-14, width=8)
    '11110010'

    """
    pass


@binary_expose.implements(np.bitwise_or, func_name="bitwise_or")
def _impl_ufunc_bitwise_or(x1, x2):
    """
    Computes the bitwise OR of the underlying binary representation of each element x1_i of the
    input array x1 with the respective element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('int64(int64, int64)')
    ... def array_api_bitwise_or(x1, x2):
    ...     return array_api.bitwise_or(x1, x2)
    >>> array_api_bitwise_or(13, 16).execute()
    29
    >>> np.binary_repr(29)
    '11101'
    >>> array_api_bitwise_or(32, 2).execute()
    34

    """
    pass


@binary_expose.implements(np.right_shift, func_name="bitwise_right_shift")
def _impl_ufunc_bitwise_right_shift(x1, x2):
    """
    Shifts the bits of each element x1_i of the input array x1 to the right by appending x2_i
    (i.e., the respective element in the input array x2) zeros to the right of x1_i.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('int64(int64, int64)')
    ... def array_api_bitwise_right_shift(x1, x2):
    ...     return array_api.bitwise_right_shift(x1, x2)
    >>> array_api_bitwise_right_shift(10, 1).execute()
    5
    >>> np.binary_repr(5)
    '101'
    """
    pass


@binary_expose.implements(np.bitwise_xor, func_name="bitwise_xor")
def _impl_ufunc_bitwise_xor(x1, x2):
    """
    Computes the bitwise XOR of the underlying binary representation of each element x1_i of the
    input array x1 with the respective element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('int64(int64, int64)')
    ... def array_api_bitwise_xor(x1, x2):
    ...     return array_api.bitwise_xor(x1, x2)
    >>> array_api_bitwise_xor(13, 17).execute()
    28
    >>> np.binary_repr(28)
    '11100'
    """
    pass


@unary_expose.implements(np.ceil, func_name="ceil", dtype=types.double)
def _impl_ufunc_ceil(x):
    """
    Rounds each element x_i of the input array x to the smallest (i.e., closest to -infinity)
    integer-valued number that is not less than x_i.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_ceil(x):
    ...     return array_api.ceil(x)
    >>> arr = np.array([-1.7, -1.5, -0.2, 0.2, 1.5, 1.7, 2.0])
    >>> array_api_ceil(arr).execute()
    array([-1., -1., -0.,  1.,  2.,  2.,  2.], dtype=float32)

    """
    pass


@unary_expose.implements(np.conj, func_name="conj")
def _impl_ufunc_conj(x):
    """
    Returns the complex conjugate for each element x_i of the input array
    """
    pass


@unary_expose.implements(np.cos, func_name="cos")
def _impl_ufunc_cos(x):
    """Calculates an implementation-dependent approximation to the cosine, having domain
    (-infinity, +infinity) and codomain [-1, +1], for each element x_i of the input
    array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_cos(x):
    ...     return array_api.cos(x)
    >>> array_api_cos(np.array([0, np.pi/2, np.pi])).execute()
    array([ 1.000000e+00,  6.123234e-17, -1.000000e+00], dtype=float32)

    """
    pass


@unary_expose.implements(np.cosh, func_name="cosh")
def _impl_ufunc_cosh(x):
    """Calculates an implementation-dependent approximation to the hyperbolic cosine, having
    domain.

    [-infinity, +infinity] and codomain [-infinity, +infinity], for each
    element x_i in the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double(double)')
    ... def array_api_cosh(x):
    ...     return array_api.cosh(x)
    >>> array_api_cosh(0.0).execute()
    1.0

    """
    pass


@binary_expose.implements(np.divide, func_name="divide")
def _impl_ufunc_divide(x1, x2):
    """
    Calculates the division for each element x1_i of the input array x1 with the respective
    element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double(double, double)')
    ... def array_api_divide(x1, x2):
    ...     return array_api.divide(x1, x2)
    >>> array_api_divide(2.0, 4.0).execute()
    0.5
    """
    pass


@binary_expose.implements(np.equal, func_name="equal", dtype=typesystem.boolean8)
def _impl_ufunc_equal(x1, x2):
    """
    Computes the truth value of x1_i == x2_i for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('bool[](int64[], int64[])')
    ... def array_api_equal(x1, x2):
    ...     return array_api.equal(x1, x2)
    >>> array_api_equal([0, 1, 3], np.arange(3)).execute()
    array([ True,  True, False])
    """
    pass


@unary_expose.implements(np.exp, func_name="exp")
def _impl_ufunc_exp(x):
    """Calculates an implementation-dependent approximation to the exponential function, having
    domain.

    [-infinity, +infinity] and codomain [+0, +infinity], for each
    element x_i of the input array x (e raised to the power of x_i,
    where e is the base of the natural logarithm).

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_exp(x):
    ...     return array_api.exp(x)
    >>> array_api_exp(np.array([1.0, np.pi, np.e])).execute()
    array([ 2.7182817, 23.140692 , 15.154263 ], dtype=float32)

    """
    pass


@unary_expose.implements(np.expm1, func_name="expm1")
def _impl_ufunc_expm1(x):
    """Calculates an implementation-dependent approximation to exp(x)-1, having domain [-infinity,

    +infinity] and codomain [-1, +infinity], for each element x_i of the
    input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_expm1(x):
    ...     return array_api.expm1(x)
    >>> array_api_expm1(np.array([1.0, np.pi, np.e])).execute()
    array([ 1.7182819, 22.140692 , 14.154263 ], dtype=float32)

    """
    pass


@unary_expose.implements(np.floor, func_name="floor", dtype=types.double)
def _impl_ufunc_floor(x):
    """
    Rounds each element x_i of the input array x to the greatest (i.e., closest to +infinity)
    integer-valued number that is not greater than x_i.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_floor(x):
    ...     return array_api.floor(x)
    >>> array_api_floor(np.array([1.0, np.pi, np.e])).execute()
    array([1., 3., 2.], dtype=float32)

    """
    pass


@binary_expose.implements(np.floor_divide, func_name="floor_divide")
def _impl_ufunc_floor_divide(x1, x2):
    """
    Rounds the result of dividing each element x1_i of the input array x1 by the respective
    element x2_i of the input array x2 to the greatest (i.e., closest to +infinity) integer-value
    number that is not greater than the division result.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('int64(int64, int64)')
    ... def array_api_bitwise_floor_divide(x1, x2):
    ...     return array_api.floor_divide(x1, x2)
    >>> array_api_bitwise_floor_divide(7, 3).execute()
    2

    """
    pass


@binary_expose.implements(np.greater, func_name="greater", dtype=typesystem.boolean8)
def _impl_ufunc_greater(x1, x2):
    """
    Computes the truth value of x1_i > x2_i for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('bool[](int64[], int64[])')
    ... def array_api_greater(x1, x2):
    ...     return array_api.greater(x1, x2)
    >>> array_api_greater([4,2],[2,2]).execute()
    array([ True, False])

    """
    pass


@binary_expose.implements(
    np.greater_equal, func_name="greater_equal", dtype=typesystem.boolean8
)
def _impl_ufunc_greater_equal(x1, x2):
    """
    Computes the truth value of x1_i >= x2_i for each element x1_i of the
    input array x1 with the respective element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('bool[](int64[], int64[])')
    ... def array_api_greater_equal(x1, x2):
    ...     return array_api.greater_equal(x1, x2)
    >>> array_api_greater_equal([4, 2, 1], [2, 2, 2]).execute()
    array([ True,  True, False])

    """
    pass


@unary_expose.not_implemented(func_name="imag")
def _impl_ufunc_imag(x):
    """
    Returns the imaginary component of a complex number for each element x_i of
    the input array x.
    """


@unary_expose.implements(np.isfinite, func_name="isfinite", dtype=typesystem.boolean8)
def _impl_ufunc_isfinite(x):
    """
    Tests each element x_i of the input array x to determine if finite (i.e., not NaN and not
    equal to positive or negative infinity).

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('bool[](double[])')
    ... def array_api_isfinite(x):
    ...     return array_api.isfinite(x)
    >>> array_api_isfinite([1.0, 0.0]).execute()
    array([ True,  True])

    """
    pass


@unary_expose.implements(np.isinf, func_name="isinf", dtype=typesystem.boolean8)
def _impl_ufunc_isinf(x):
    """
    Tests each element x_i of the input array x to determine if equal to positive or negative
    infinity.
    """
    pass


@unary_expose.implements(np.isnan, func_name="isnan", dtype=typesystem.boolean8)
def _impl_ufunc_isnan(x):
    """
    Tests each element x_i of the input array x to determine whether the element is NaN.
    """
    pass


@binary_expose.implements(np.less, func_name="less", dtype=typesystem.boolean8)
def _impl_ufunc_less(x1, x2):
    """
    Computes the truth value of x1_i < x2_i for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('bool[](int64[], int64[])')
    ... def array_api_less(x1, x2):
    ...     return array_api.less(x1, x2)
    >>> array_api_less([4, 2, 1], [2, 2, 2]).execute()
    array([False, False,  True])

    """
    pass


@binary_expose.implements(
    np.less_equal, func_name="less_equal", dtype=typesystem.boolean8
)
def _impl_ufunc_less_equal(x1, x2):
    """
    Computes the truth value of x1_i <= x2_i for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('bool[](int64[], int64[])')
    ... def array_api_less_equal(x1, x2):
    ...     return array_api.less_equal(x1, x2)
    >>> array_api_less_equal([4, 2, 1], [2, 2, 2]).execute()
    array([False,  True,  True])

    """
    pass


@unary_expose.implements(np.log, func_name="log")
def _impl_ufunc_log(x):
    """
    Calculates an implementation-dependent approximation to the natural (base e) logarithm, having
    domain [0, +infinity] and codomain [-infinity, +infinity], for each element x_i of the input
    array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_log(x):
    ...     return array_api.log(x)
    >>> array_api_log([1.0, np.e, np.e**2, 0.0]).execute()
    array([  0.,   1.,   2., -inf], dtype=float32)

    """
    pass


@unary_expose.implements(np.log1p, func_name="log1p")
def _impl_ufunc_log1p(x):
    """
    Calculates an implementation-dependent approximation to log(1+x), where log refers to the
    natural (base e) logarithm, having domain [-1, +infinity] and codomain [-infinity, +infinity],
    for each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_log1p(x):
    ...     return array_api.log1p(x)
    >>> array_api_log1p([np.e, 8.0]).execute()
    array([1.3132616, 2.1972246], dtype=float32)

    """
    pass


@unary_expose.implements(np.log2, func_name="log2")
def _impl_ufunc_log2(x):
    """Calculates an implementation-dependent approximation to the base 2 logarithm, having
    domain. [0, +infinity] and codomain [-infinity, +infinity], for each element x_i
    of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_log2(x):
    ...     return array_api.log2(x)
    >>> array_api_log2([np.e, 8.0]).execute()
    array([1.442695, 3.      ], dtype=float32)

    """
    pass


@unary_expose.implements(np.log10, func_name="log10")
def _impl_ufunc_log10(x):
    """Calculates an implementation-dependent approximation to the base 10 logarithm, having
    domain. [0, +infinity] and codomain [-infinity, +infinity], for each element x_i
    of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_log10(x):
    ...     return array_api.log10(x)
    >>> array_api_log10([0.01, -3.0]).execute()
    array([-2., nan], dtype=float32)

    """
    pass


@binary_expose.implements(np.logaddexp, func_name="logaddexp")
def _impl_ufunc_logaddexp(x1, x2):
    """
    Calculates the logarithm of the sum of exponentiations ``log(exp(x1) + exp(x2))`` for each
    element x1_i of the input array x1 with the respective element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double(double, double)')
    ... def array_api_logaddexp(x1, x2):
    ...     return array_api.logaddexp(x1, x2)
    >>> prob1 = np.log(1e-50)
    >>> prob2 = np.log(2.5e-50)
    >>> array_api_logaddexp(prob1, prob2).execute()
    -113.876495

    """
    pass


@binary_expose.implements(
    np.logical_and, func_name="logical_and", dtype=typesystem.boolean8
)
def _impl_ufunc_logical_and(x1, x2):
    """
    Computes the logical AND for each element x1_i of the input array x1 with the respective
    element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(
    np.logical_or, func_name="logical_or", dtype=typesystem.boolean8
)
def _impl_ufunc_logical_or(x1, x2):
    """
    Computes the logical OR for each element x1_i of the input array x1 with the respective
    element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(
    np.logical_xor, func_name="logical_xor", dtype=typesystem.boolean8
)
def _impl_ufunc_logical_xor(x1, x2):
    """
    "Computes the logical XOR for each element x1_i of the input array x1 with the respective
    element x2_i of the input array x2.
    """
    pass


@unary_expose.implements(np.logical_not, dtype=typesystem.boolean8)
def _impl_ufunc_logical_not(x):
    """
    Computes the logical NOT for each element x_i of the input array x.
    """
    pass


@binary_expose.implements(np.multiply, func_name="multiply")
def _impl_ufunc_multiply(x1, x2):
    """
    Calculates the product for each element x1_i of the input array x1 with the respective element
    x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double(double, double)')
    ... def array_api_multiply(x1, x2):
    ...     return array_api.multiply(x1, x2)
    >>> array_api_multiply(2.0, 4.0).execute()
    8.0

    """
    pass


@unary_expose.implements(np.negative, func_name="negative")
def _impl_ufunc_negative(x):
    """Computes the numerical negative of each element x_i (i.e., y_i = -x_i) of the
    input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_negative(x):
    ...     return array_api.negative(x)
    >>> array_api_negative([1., -1.]).execute()
    array([-1.,  1.], dtype=float32)

    """
    pass


@binary_expose.implements(
    np.not_equal, func_name="not_equal", dtype=typesystem.boolean8
)
def _impl_ufunc_not_equal(x1, x2):
    """
    Computes the truth value of x1_i != x2_i for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.
    """
    pass


@unary_expose.implements(np.positive, func_name="positive")
def _impl_ufunc_positive(x):
    """Computes the numerical positive of each element x_i (i.e., y_i = +x_i) of the
    input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_positive(x):
    ...     return array_api.positive(x)
    >>> array_api_positive([1., -1.]).execute()
    array([ 1., -1.], dtype=float32)

    """
    pass


@binary_expose.implements(np.power, func_name="pow")
def _impl_ufunc_pow(x1, x2):
    """
    Calculates an implementation-dependent approximation of exponentiation by raising each element
    x1_i (the base) of the input array x1 to the power of x2_i (the exponent), where x2_i is the
    corresponding element of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('int64[](int64[], int64)')
    ... def array_api_pow(x1, x2):
    ...     return array_api.pow(x1, x2)
    >>> x1 = np.arange(6)
    >>> array_api_pow(x1, 3).execute()
    array([  0,   1,   8,  27,  64, 125])

    """
    pass


@unary_expose.not_implemented(func_name="real")
def _impl_ufunc_real(x):
    """
    Returns the real component of a complex number for each element x_i of the
    input array x.
    """
    pass


@binary_expose.implements(np.remainder, func_name="remainder")
def _impl_ufunc_remainder(x1, x2):
    """
    Returns the remainder of division for each element x1_i of the input array x1 and the
    respective element x2_i of the input array x2.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('int64[](int64[], int64[])')
    ... def array_api_remainder(x1, x2):
    ...     return array_api.remainder(x1, x2)
    >>> array_api_remainder([4, 7], [2, 3]).execute()
    array([0, 1])

    """
    pass


@unary_expose.implements(np.round, func_name="round")
def _impl_ufunc_round(x):
    """
    Rounds each element x_i of the input array x to the nearest integer-valued number.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_round(x):
    ...     return array_api.round(x)
    >>> array_api_round([0.37, 1.64]).execute()
    array([0., 2.], dtype=float32)

    """
    pass


@unary_expose.implements(np.sign, func_name="sign")
def _impl_ufunc_sign(x):
    """
    Returns an indication of the sign of a number for each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_sign(x):
    ...     return array_api.sign(x)
    >>> array_api_sign([-5., 4.5]).execute()
    array([-1.,  1.], dtype=float32)

    """
    pass


@unary_expose.implements(np.sin, func_name="sin")
def _impl_ufunc_sin(x):
    """
    Calculates an implementation-dependent approximation to the sine for each
    element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double(double)')
    ... def array_api_sin(x):
    ...     return array_api.sin(x)
    >>> array_api_sin(np.pi/2.).execute()
    1.0

    """
    pass


@unary_expose.implements(np.sinh, func_name="sinh")
def _impl_ufunc_sinh(x):
    """
    Calculates an implementation-dependent approximation to the hyperbolic sine
    for each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double(double)')
    ... def array_api_sinh(x):
    ...     return array_api.sinh(x)
    >>> array_api_sinh(0).execute()
    0.0
    >>> array_api_sinh(np.pi).execute()
    11.548739

    """
    pass


@unary_expose.implements(np.square, func_name="square")
def _impl_ufunc_square(x):
    """
    Squares (x_i * x_i) each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double(double)')
    ... def array_api_square(x):
    ...     return array_api.square(x)
    >>> array_api_square(4.0).execute()
    16.0

    """
    pass


@unary_expose.implements(np.sqrt, func_name="sqrt")
def _impl_ufunc_sqrt(x):
    """
    Calculates the principal square root for each element x_i of the input
    array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double(double)')
    ... def array_api_sqrt(x):
    ...     return array_api.sqrt(x)
    >>> array_api_sqrt(4.0).execute()
    2.0

    """
    pass


@binary_expose.implements(np.subtract, func_name="subtract")
def _impl_ufunc_subtract(x1, x2):
    """
    Calculates the difference for each element x1_i of the input array x1 with
    the respective element x2_i of the input array x2

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('int64(int64, int64)')
    ... def array_api_subtract(x1, x2):
    ...     return array_api.subtract(x1, x2)
    >>> array_api_subtract(5, 3).execute()
    2

    """
    pass


@unary_expose.implements(np.tan, func_name="tan")
def _impl_ufunc_tan(x):
    """
    Calculates an implementation-dependent approximation to the tangent for
    each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_tan(x):
    ...     return array_api.tan(x)
    >>> array_api_tan([-np.pi, np.pi/2, np.pi]).execute()
    array([ 1.2246469e-16,  1.6331239e+16, -1.2246469e-16], dtype=float32)

    """
    pass


@unary_expose.implements(np.tanh, func_name="tanh")
def _impl_ufunc_tanh(x):
    """
    Calculates an implementation-dependent approximation to the hyperbolic
    tangent for each element x_i of the input array x.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double[](double[])')
    ... def array_api_tanh(x):
    ...     return array_api.tanh(x)
    >>> array_api_tanh([0.0, np.pi/2, np.pi]).execute()
    array([0.        , 0.91715235, 0.9962721 ], dtype=float32)

    """
    pass


@unary_expose.implements(np.trunc, func_name="trunc", dtype=types.double)
def _impl_ufunc_trunc(x):
    """
    Rounds each element x_i of the input array x to the nearest integer-valued
    number that is closer to zero than x_i.

    Examples
    --------
    IGNORE:
    >>> from rbc.heavydb import global_heavydb_singleton
    >>> heavydb = next(global_heavydb_singleton)
    >>> heavydb.unregister()

    IGNORE

    >>> from rbc.stdlib import array_api
    >>> @heavydb('double(double)')
    ... def array_api_trunc(x):
    ...     return array_api.trunc(x)
    >>> array_api_trunc(np.pi).execute()
    3.0

    """
    pass


###############


@binary_expose.implements(np.copysign, api=API.NUMPY_API)
def _impl_ufunc_copysign(x1, x2):
    pass


@binary_expose.implements(np.logaddexp2, api=API.NUMPY_API)
def _impl_ufunc_logaddexp2(x1, x2):
    pass


@binary_expose.implements(np.true_divide, func_name="true_divide", api=API.NUMPY_API)
def _impl_ufunc_true_divide(x1, x2):
    pass


@binary_expose.implements(np.power, func_name="power", api=API.NUMPY_API)
def _impl_ufunc_power(x1, x2):
    pass


@binary_expose.not_implemented("float_power")  # not supported by Numba
def _impl_ufunc_float_power(x1, x2):
    pass


@binary_expose.implements(np.mod, func_name="mod", api=API.NUMPY_API)
def _impl_ufunc_mod(x1, x2):
    pass


@binary_expose.implements(np.fmod, api=API.NUMPY_API)
def _impl_ufunc_fmod(x1, x2):
    pass


@binary_expose.not_implemented("divmod")  # not supported by Numba
def _impl_ufunc_divmod(x1, x2):
    pass


@binary_expose.implements(np.gcd, api=API.NUMPY_API)
def _impl_ufunc_gcd(x1, x2):
    pass


@binary_expose.implements(np.lcm, api=API.NUMPY_API)
def _impl_ufunc_lcm(x1, x2):
    pass


# Bit-twiddling functions
@binary_expose.implements(np.left_shift, api=API.NUMPY_API)
def _impl_ufunc_left_shift(x1, x2):
    pass


@binary_expose.implements(np.right_shift, api=API.NUMPY_API)
def _impl_ufunc_right_shift(x1, x2):
    pass


# trigonometric functions
@binary_expose.implements(np.arctan2, api=API.NUMPY_API)
def _impl_ufunc_arctan2(x1, x2):
    pass


@binary_expose.implements(np.hypot, api=API.NUMPY_API)
def _impl_ufunc_hypot(x1, x2):
    pass


# Comparison functions

@binary_expose.implements(np.maximum, api=API.NUMPY_API)
def _impl_ufunc_maximum(x1, x2):
    pass


@binary_expose.implements(np.minimum, api=API.NUMPY_API)
def _impl_ufunc_minimum(x1, x2):
    pass


@binary_expose.implements(np.fmax, api=API.NUMPY_API)
def _impl_ufunc_fmax(x1, x2):
    pass


@binary_expose.implements(np.fmin, api=API.NUMPY_API)
def _impl_ufunc_fmin(x1, x2):
    pass


# Floating functions
@binary_expose.implements(np.nextafter, api=API.NUMPY_API)
def _impl_ufunc_nextafter(x1, x2):
    pass


@binary_expose.implements(np.ldexp, api=API.NUMPY_API)
def _impl_ufunc_ldexp(x1, x2):
    pass


##################################################################


@unary_expose.implements(np.absolute, api=API.NUMPY_API)
def _impl_ufunc_absolute(x):
    pass


@unary_expose.implements(np.rint, api=API.NUMPY_API)
def _impl_ufunc_rint(x):
    pass


@unary_expose.implements(np.conjugate, api=API.NUMPY_API)
def _impl_ufunc_conjugate(x):
    pass


@unary_expose.implements(np.exp2, api=API.NUMPY_API)
def _impl_ufunc_exp2(x):
    pass


# @unary_expose.implements(np.cbrt)  # not supported by numba
@unary_expose.not_implemented("cbrt")
def _impl_ufunc_cbrt(x):
    pass


@unary_expose.implements(np.reciprocal, api=API.NUMPY_API)
def _impl_ufunc_reciprocal(x):
    pass


# Bit-twiddling functions
@unary_expose.implements(np.invert, api=API.NUMPY_API)
def _impl_ufunc_invert(x):
    pass


# trigonometric functions


@unary_expose.implements(np.arcsin, api=API.NUMPY_API)
def _impl_ufunc_arcsin(x):
    pass


@unary_expose.implements(np.arccos, api=API.NUMPY_API)
def _impl_ufunc_arccos(x):
    pass


@unary_expose.implements(np.arctan, api=API.NUMPY_API)
def _impl_ufunc_arctan(x):
    pass


@unary_expose.implements(np.arcsinh, api=API.NUMPY_API)
def _impl_ufunc_arcsinh(x):
    pass


@unary_expose.implements(np.arccosh, api=API.NUMPY_API)
def _impl_ufunc_arccosh(x):
    pass


@unary_expose.implements(np.arctanh, api=API.NUMPY_API)
def _impl_ufunc_arctanh(x):
    pass


@unary_expose.implements(np.degrees, api=API.NUMPY_API)
def _impl_ufunc_degrees(x):
    pass


@unary_expose.implements(np.radians, api=API.NUMPY_API)
def _impl_ufunc_radians(x):
    pass


@unary_expose.implements(np.deg2rad, api=API.NUMPY_API)
def _impl_ufunc_deg2rad(x):
    pass


@unary_expose.implements(np.rad2deg, api=API.NUMPY_API)
def _impl_ufunc_rad2deg(x):
    pass


# Floating functions


@unary_expose.implements(np.fabs, dtype=types.double, api=API.NUMPY_API)
def _impl_ufunc_fabs(x):
    pass


# not supported?
# @unary_expose.implements(np.isnat, dtype=types.int8)
@unary_expose.not_implemented("isnat", api=API.NUMPY_API)
def _impl_ufunc_isnat(x):
    pass


# issue 152:
@unary_expose.implements(np.signbit, dtype=typesystem.boolean8, api=API.NUMPY_API)
def _impl_ufunc_signbit(x):
    pass


@unary_expose.implements(np.spacing, dtype=types.double, api=API.NUMPY_API)
def _impl_ufunc_spacing(x):
    pass


@expose.implements("heaviside", api=API.NUMPY_API)
def _impl_heaviside(x1, x2):
    nb_dtype = types.double
    typA = determine_input_type(x1)
    typB = determine_input_type(x2)
    if isinstance(x1, ArrayPointer):

        def impl(x1, x2):
            sz = len(x1)
            r = Array(sz, nb_dtype)
            for i in range(sz):
                r[i] = heaviside(x1[i], x2)  # noqa: F821
            return r

        return impl
    else:

        def impl(x1, x2):
            if typA(x1) < 0:
                return nb_dtype(0)
            elif typA(x1) == 0:
                return nb_dtype(typB(x2))
            else:
                return nb_dtype(1)

        return impl
